"""
Provide command line interface for htmlcrypt tool
"""

import argparse
import pathlib
from getpass import getpass
from .encrypt import encrypt_file
from .decrypt import write_decrypt_file


# pylint: disable=too-many-branches
def main() -> None:
    """Use sys.argv arguments to specify file for encryption"""
    parser = argparse.ArgumentParser(
        prog="python -m htmlcrypt" if __name__ == "__main__" else None,
        description="Encrypt HTML file with password",
    )
    parser.add_argument(
        "-p",
        "--password",
        help="""Password for encrypting/decrypting.
        Will ask for password from stdin if not specified""",
    )
    input_file_arg = parser.add_argument(
        "file", type=pathlib.Path, help="File to encrypt"
    )
    encrypted_file_arg = parser.add_argument(
        "-e", "--encrypted", type=pathlib.Path, help="Encrypted filename"
    )
    decrypt_file_arg = parser.add_argument(
        "-d",
        "--decrypt",
        type=pathlib.Path,
        help="""Location to generate the decryption .html file.
        Relative path to encrypted file will be injected into file.""",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_const",
        const=True,
        help="Overwrite existing output file",
    )

    args = parser.parse_args()

    encrypted_file_arg.default = (
        str(args.decrypt.with_suffix(args.decrypt.suffix + ".content.encrypted"))
        if args.decrypt
        else str(args.file.with_suffix(args.file.suffix + ".encrypted"))
    )
    args = parser.parse_args()

    file_set = {args.file.resolve(), args.encrypted.resolve()}
    if len(file_set) != 2:
        parser.error(
            "You cannot supply the same file as input file and --encrypted argument"
        )

    if args.decrypt:
        file_set.add(args.decrypt.resolve())
        if len(file_set) != 3:
            parser.error(
                "The decryption file cannot be the same as the input or encrypted file"
            )

    if args.password is not None:
        password = args.password
    else:
        while True:
            password = getpass("Password: ")
            if password == getpass("Confirm: "):
                break
            print("Passwords did not match. Try again")

    input_file_arg.type = argparse.FileType("rb")
    encrypted_file_arg.type = argparse.FileType("xb" if not args.force else "wb")
    decrypt_file_arg.type = argparse.FileType("x" if not args.force else "w")
    args = parser.parse_args()

    print(f"Encrypting {args.file.name} to {args.encrypted.name}")

    with args.file as input_file:
        with args.encrypted as output_file:
            encrypt_file(password, input_file, output_file)

    if args.decrypt:
        print(f"Writing decrypt file to {args.decrypt.name}")
        write_decrypt_file(args.decrypt, args.encrypted.name)


if __name__ == "__main__":
    main()



"""
Define methods for encrypting file with password
"""

import typing
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Hash import SHA256


# pylint: disable=too-few-public-methods
class _SupportsRead(typing.Protocol):
    # pylint: disable=missing-function-docstring
    def read(self, amount: int) -> bytes:
        ...


class _SupportsEncrypt(typing.Protocol):
    # pylint: disable=missing-function-docstring
    def encrypt(self, data: bytes) -> bytes:
        ...

    def encrypt_and_digest(self, data: bytes) -> typing.Tuple[bytes, bytes]:
        ...


def _create_salt(rand: _SupportsRead) -> bytes:
    "Create the salt using the default config using the provided random source"
    return rand.read(32)


def _derive_key(password: str, salt: bytes) -> bytes:
    "Derive the encryption key using PBKDF2 from password and salt"
    return PBKDF2(
        password,
        salt,
        count=100_000,
        dkLen=32,
        hmac_hash_module=SHA256,
    )


def _create_nonce(rand: _SupportsRead) -> bytes:
    "Create the nonce using the default config using the provided random source"
    return rand.read(16)


def _get_cipher(key: bytes, nonce: bytes) -> _SupportsEncrypt:
    "Create the AES Cipher"
    return typing.cast(_SupportsEncrypt, AES.new(key, AES.MODE_GCM, nonce=nonce))


def _encrypt_file_to_file(
    cipher: _SupportsEncrypt, inputfile: typing.BinaryIO, outputfile: typing.BinaryIO
) -> bytes:
    "Use the proved cipher to encrypt the content of inputfile and write to outputfile"
    while True:
        chunk = inputfile.read(1024)
        if not chunk:
            break
        outputfile.write(cipher.encrypt(chunk))
    encrypted, digest = cipher.encrypt_and_digest(b"")
    outputfile.write(encrypted)
    return digest


def encrypt_file(
    password: str, input_file: typing.BinaryIO, output_file: typing.BinaryIO
) -> None:
    "Encrypt file with password and write to output"
    rand = Random.new()

    salt = _create_salt(rand)
    key = _derive_key(password, salt)
    nonce = _create_nonce(rand)
    cipher = _get_cipher(key, nonce)

    output_file.write(salt)
    output_file.write(nonce)
    digest = _encrypt_file_to_file(cipher, input_file, output_file)
    output_file.write(digest)



"""
Tool for encrypting HTML files for later decryption with JavaScript.

Depends on pycryptodome.
"""

__all__ = [
    "encrypt_file",
    "write_decrypt_file",
]

from .encrypt import encrypt_file
from .decrypt import write_decrypt_file



# pylint: disable=missing-module-docstring,missing-function-docstring,duplicate-code

import pathlib
import sys
from typing import Optional
import pytest
import htmlcrypt.__main__


def create_test_file(path: pathlib.Path) -> None:
    assert not path.exists()
    path.write_text("Test content file")
    assert path.is_file()


def run_with_args(*args: str) -> None:
    del sys.argv[1:]
    sys.argv.extend(args)
    htmlcrypt.__main__.main()


def run_with_files(
    inputfile: pathlib.Path,
    encrypted: Optional[pathlib.Path] = None,
    decrypt: Optional[pathlib.Path] = None,
    password: str = "password",
) -> None:
    args = [str(inputfile), "-p", password]
    if encrypted is not None:
        args.extend(["-e", str(encrypted)])
    if decrypt is not None:
        args.extend(["-d", str(decrypt)])
    run_with_args(*args)
    if encrypted is not None:
        assert encrypted.is_file()
    else:
        if decrypt is not None:
            assert decrypt.with_suffix(decrypt.suffix + ".content.encrypted")
        else:
            assert inputfile.with_suffix(inputfile.suffix + ".encrypted").is_file()
    if decrypt is not None:
        assert decrypt.is_file()


def test_missing_file(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    assert not input_file_path.exists()
    with pytest.raises(SystemExit):
        run_with_args(str(input_file_path), "-p", "password")


def test_single_arg(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    create_test_file(input_file_path)
    run_with_files(input_file_path)


def test_specify_encrypted_filename(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    encrypted_file_path = tmp_path / "encrypted_input.encrypted"
    create_test_file(input_file_path)
    run_with_files(input_file_path, encrypted=encrypted_file_path)


def test_specify_encrypted_filename_with_subfolder(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    encrypted_file_path = tmp_path / "subfolder" / "encrypted_input.encrypted"
    assert not encrypted_file_path.parent.exists()
    encrypted_file_path.parent.mkdir()
    assert encrypted_file_path.parent.is_dir()
    create_test_file(input_file_path)
    run_with_files(input_file_path, encrypted=encrypted_file_path)


def test_specify_decryption_filename(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    create_test_file(input_file_path)
    decrypt_file_path = tmp_path / "decrypt.html"
    assert not decrypt_file_path.exists()
    run_with_files(input_file_path, decrypt=decrypt_file_path)


def test_specify_all_filenames(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    create_test_file(input_file_path)
    decrypt_file_path = tmp_path / "decrypt.html"
    encrypted_file_path = tmp_path / "encrypted_content"
    assert not decrypt_file_path.exists()
    assert not encrypted_file_path.exists()
    run_with_files(
        input_file_path, encrypted=encrypted_file_path, decrypt=decrypt_file_path
    )


def test_empty_password(tmp_path: pathlib.Path) -> None:
    input_file_path = tmp_path / "input.txt"
    create_test_file(input_file_path)
    run_with_files(input_file_path, password="")



"""
Define methods for generating the HTML file used to decrypt the content
"""

import typing
import re
import os.path
import shutil
from importlib.resources import files


def write_decrypt_file(decrypt_file: typing.TextIO, encrypted_file_path: str) -> None:
    """
    Write HTML to `decrypt_file` argument that will prompt to decrypt file from
    `encrypted_file_path` with password.
    Generates a relative path.
    """
    rex = re.compile(r"const encryptedFileName = 'content.html.encrypted';")
    encrypted_file_rel_path = os.path.relpath(
        encrypted_file_path, os.path.dirname(decrypt_file.name)
    )
    value = f"const encryptedFileName = '{encrypted_file_rel_path}';"
    with files(__package__).joinpath("decrypt.html").open("r") as decrypt_template:
        for line in decrypt_template:
            newline, nreplace = rex.subn(value, line)
            decrypt_file.write(newline)
            if nreplace > 0:
                break
        else:
            assert False, "Could not find filename pattern in decrypt.html template"
        # Copy remaining file content
        shutil.copyfileobj(decrypt_template, decrypt_file)
